import matplotlib.pyplot as plt
import random

# debe estar instalado matplotlib


# Función para generar una población inicial aleatoria
def generar_poblacion(num_individuos, num_ciudades):
    poblacion = []
    for _ in range(num_individuos):
        individuo = list(range(num_ciudades))
        random.shuffle(individuo)
        poblacion.append(individuo)
    return poblacion


# Función para evaluar la aptitud de un individuo (distancia total del recorrido)
def calcular_aptitud(individuo, distancias):
    distancia_total = 0
    for i in range(len(individuo) - 1):
        ciudad_actual = individuo[i]
        siguiente_ciudad = individuo[i + 1]
        distancia_total += distancias[ciudad_actual][siguiente_ciudad]
    distancia_total += distancias[individuo[-1]][individuo[0]]
    return distancia_total


# Función para seleccionar individuos para la reproducción (torneo binario)
def seleccion_torneo(poblacion, distancias):
    seleccionados = []
    for _ in range(len(poblacion)):
        torneo = random.sample(poblacion, 2)
        aptitud_torneo = [
            calcular_aptitud(individuo, distancias) for individuo in torneo
        ]
        seleccionado = torneo[aptitud_torneo.index(min(aptitud_torneo))]
        seleccionados.append(seleccionado)
    return seleccionados


# Función para realizar el cruce de dos padres para producir un hijo
def cruzar(padre1, padre2):
    punto_cruce = random.randint(0, len(padre1) - 1)
    hijo = padre1[:punto_cruce] + [
        gen for gen in padre2 if gen not in padre1[:punto_cruce]
    ]
    return hijo


# Función para aplicar mutaciones en la población
def mutar(individuo, probabilidad_mutacion):
    if random.random() < probabilidad_mutacion:
        indices = random.sample(range(len(individuo)), 2)
        individuo[indices[0]], individuo[indices[1]] = (
            individuo[indices[1]],
            individuo[indices[0]],
        )
    return individuo


# Función para generar distancias aleatorias entre ciudades y sus coordenadas tridimensionales
def generar_distancias(num_ciudades):
    distancias = [[0] * num_ciudades for _ in range(num_ciudades)]
    coordenadas = [
        (random.uniform(0, 100), random.uniform(0, 100), random.uniform(0, 100))
        for _ in range(num_ciudades)
    ]

    for i in range(num_ciudades):
        for j in range(i + 1, num_ciudades):
            distancias[i][j] = distancias[j][i] = (
                sum((x - y) ** 2 for x, y in zip(coordenadas[i], coordenadas[j])) ** 0.5
            )

    return distancias, coordenadas


def visualizar_camino(camino, coordenadas, mejor_distancia):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    x = [coordenadas[i][0] for i in camino]
    y = [coordenadas[i][1] for i in camino]
    z = [coordenadas[i][2] for i in camino]
    ax.plot(x, y, z, marker="o", label="Camino")
    ax.scatter(x[0], y[0], z[0], color="green", s=100, label="Inicio")
    ax.text(x[0], y[0], z[0], "Inicio", fontsize=10)
    ax.scatter(x[-1], y[-1], z[-1], color="red", s=100, label="Fin")
    ax.text(x[-1], y[-1], z[-1], "Fin", fontsize=10)
    for i, (xi, yi, zi) in enumerate(zip(x, y, z)):
        ax.text(xi, yi, zi, str(camino[i]), fontsize=8)

    ax.set_title(f"Mejor Camino Encontrado\nDistancia: {mejor_distancia:.2f}")
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    ax.legend()
    plt.show()


def algoritmo_genetico(
    num_generaciones, num_ciudades, num_individuos, probabilidad_mutacion, distancias
):
    poblacion = generar_poblacion(num_individuos, num_ciudades)
    mejor_solucion_historial = []
    mejor_distancia_historial = []

    for generacion in range(num_generaciones):
        poblacion = sorted(poblacion, key=lambda x: calcular_aptitud(x, distancias))
        mejor_individuo = poblacion[0]
        mejor_distancia = calcular_aptitud(mejor_individuo, distancias)
        # Almacenar el mejor individuo y su distancia en cada generación
        mejor_solucion_historial.append(mejor_individuo)
        mejor_distancia_historial.append(mejor_distancia)

        seleccionados = seleccion_torneo(poblacion, distancias)

        nueva_poblacion = []
        for i in range(0, len(seleccionados), 2):
            padre1, padre2 = seleccionados[i], seleccionados[i + 1]
            hijo1 = cruzar(padre1, padre2)
            hijo2 = cruzar(padre2, padre1)
            hijo1 = mutar(hijo1, probabilidad_mutacion)
            hijo2 = mutar(hijo2, probabilidad_mutacion)
            nueva_poblacion.extend([hijo1, hijo2])

        poblacion = nueva_poblacion

    mejor_solucion = poblacion[0]
    mejor_distancia = calcular_aptitud(mejor_solucion, distancias)

    # Visualizar el proceso del algoritmo
    visualizar_proceso(
        mejor_distancia_historial, mejor_solucion, coordenadas, mejor_distancia
    )

    return mejor_solucion, mejor_distancia


def visualizar_proceso(
    mejor_distancia_historial, mejor_solucion, coordenadas, mejor_distancia
):
    generaciones = range(len(mejor_distancia_historial))
    fig, axs = plt.subplots(1, 2, figsize=(15, 5))
    # Subgráfico 1: Evolución de la distancia
    axs[0].plot(generaciones, mejor_distancia_historial, marker="o")
    axs[0].set_title("Evolución de la Distancia en Cada Generación")
    axs[0].set_xlabel("Generación")
    axs[0].set_ylabel("Distancia")
    # Subgráfico 2: Mejor camino encontrado
    axs[1] = fig.add_subplot(122, projection="3d")
    x = [coordenadas[i][0] for i in mejor_solucion]
    y = [coordenadas[i][1] for i in mejor_solucion]
    z = [coordenadas[i][2] for i in mejor_solucion]
    axs[1].plot(x, y, z, marker="o", label="Camino")
    axs[1].scatter(x[0], y[0], z[0], color="green", s=100, label="Inicio")
    axs[1].text(x[0], y[0], z[0], "Inicio", fontsize=10)
    axs[1].scatter(x[-1], y[-1], z[-1], color="red", s=100, label="Fin")
    axs[1].text(x[-1], y[-1], z[-1], "Fin", fontsize=10)
    for i, (xi, yi, zi) in enumerate(zip(x, y, z)):
        axs[1].text(xi, yi, zi, str(mejor_solucion[i]), fontsize=8)

    axs[1].set_title(f"Mejor Camino Encontrado\nDistancia: {mejor_distancia:.2f}")
    axs[1].set_xlabel("X")
    axs[1].set_ylabel("Y")
    axs[1].set_zlabel("Z")
    axs[1].legend()

    plt.tight_layout()
    plt.show()


# Ejemplo de uso
if __name__ == "__main__":
    # Generar un número aleatorio de ciudades o puntos en el espacio
    num_ciudades = random.randint(5, 20)
    # Generar distancias aleatorias entre las ciudades y sus coordenadas tridimensionales
    distancias, coordenadas = generar_distancias(num_ciudades)
    # parámetros del algoritmo genético
    num_generaciones = 100
    num_individuos = 50
    probabilidad_mutacion = 0.1
    # Ejecutar el algoritmo genético
    mejor_solucion, mejor_distancia = algoritmo_genetico(
        num_generaciones,
        num_ciudades,
        num_individuos,
        probabilidad_mutacion,
        distancias,
    )
    print("Mejor solución:", mejor_solucion)
    print("Mejor distancia:", mejor_distancia)
